#!/usr/bin/env python

# The command line utility.  This is called by the shell script and Windows batch file in the
# scripts directory.

import sys
from .findfile import findfile
from importlib.machinery import SourceFileLoader

args = None
# Once we parse out any global options, these will be the args tasks and options that
# are procesed by `run`.


def run():
    # Parse the global command line options. (Everything before the first task name, which we
    # assume is the first thing that does not have a dash.)
    #
    # ArgumentParser.parse_known_args looked like it did what we want but it will *skip* things
    # it doesn't recognize and grab arguments from tasks.  That makes no sense to me.  We'll
    # use our own task parser by making a "run" task to deal with the global flags.a

    def _run(help=False, verbose=False, version=False, list=False, filename='tasks.py'):
        global args

        if help:
            print('run [options] task1 [task options] task2 [task options] ...')
            print('')
            print(' -h --help      show this help message and exit')
            print(' -V --version   print version and exit')
            print(' -v --verbose   incrase verbosity')
            print(' -f --filename FILENAME')
            print('                the name of the tasks file to search for')
            return 1

        if version:
            from runtasks import version
            print('runtasks version %s' % version)
            return 1

        # Find the tasks file and load it.

        fqn = findfile(filename=filename)
        l = SourceFileLoader('tasks', fqn)
        m = l.load_module()

        # Find all of the functions in the module marked as tasks.

        tasks = {name: func._task for (name, func) in m.__dict__.items() if hasattr(func, '_task')}

        if list:
            from runtasks.lister import print_list
            print_list(tasks)
            return 1

        # Parse the task arguments.

        from runtasks.parser import parse
        parsed = parse(args, tasks)

        for (task, args) in parsed:
            if verbose == 1:
                print(task.name)
            elif verbose > 1:
                print('-' * 80)
                print('Running', task.name)
                print()

            task.call(args)

            if verbose > 1:
                print()

    from runtasks.tasks import Task
    task = Task(_run, flags={
        'V': 'version',
        'v': 'verbose'
    })

    from runtasks.parser import parse_task
    parsed, args = parse_task(task, sys.argv, [])
    task.call(parsed)




import inspect

def print_list(tasks):
    """
    Called when "--list" is passed on the command line.  Prints the list of available tasks.
    """
    names = sorted(list(tasks.keys()))
    print('Available tasks:')
    print()
    for name in names:
        task = tasks[name]
        func = task.func

        print(name)
        sig = inspect.signature(task.func)
        if sig.parameters:
            print(' ' + ' '.join(str(p) for p in sig.parameters.values()))

        doc = getattr(task.func, '__doc__', None)
        if doc:
            doc = ' ' + doc.strip().replace('\n', '\n ')
            print(doc)

        print()




import sys, inspect, logging
from collections import OrderedDict

logger = logging.getLogger('')


def parse(args, tasks):
    """
    Parses command line arguments into groups of tasks and arguments to those tasks.

    Returns a sequence of (Task, args) pairs where the arguments are an inspect.BoundArguments
    instance.
    """
    results = []

    other_names = list(tasks.keys())

    while args:
        task = tasks.get(args[0])
        if not task:
            sys.exit("I don't know what %r is" % args[0])
        boundargs, args = parse_task(task, args, other_names)

        results.append( (task, boundargs) )

    return results


def parse_task(task, cmdline, other_names):
    """
    Parses the arguments for a single task.  Raises an exception if all required arguments are
    not provided.

    Since the command line may contain multiple tasks, it stops when all parameters have been
    assigned or when unrecognized tokens are found.

    cmdline:
      An array of strings, originally from sys.argv.  The name of this task will be the first
      item.

    other_names
      Names of other defined tasks.

    Returns the tasks arguments as an inspect.BoundArguments instance and the remaining,
    unparsed arguments: (BoundArguments, remaning-cmdline)
    """
    names  = list(task.sig.parameters.keys())
    params = list(task.sig.parameters.values())

    # How many positional parameters are required?  If there are unrecognized tokens we'll
    # assume they are values for these.  For example:
    #
    #   @task test(arg1, arg2=None)
    #
    #   build test xyz --> arg1=xyz
    #
    # We only read positional parameters like this until we hit the first parameter with
    # equals or dashes ("--arg1" or "arg1=xyz", etc.).  Once we hit one of those, we'll set
    # `positional` to None to indicate so.

    positional = []
    for param in params:
        if param.default == inspect.Parameter.empty:
            positional.append(param.name)
        else:
            break

    params = OrderedDict()
    # We'll copy values from the command line (tokens) to here by parameter name.

    cmdline = cmdline[1:] # we've now copied it

    while cmdline:
        arg = cmdline[0]

        if arg.startswith('--'):
            positional = None # Now that we've seen a named item, no more positional.
            _parse_double_dash(task, cmdline, params)

        elif arg.startswith('-'):
            positional = None
            _parse_flags(task, cmdline, params)

        elif '=' in arg:
            positional = None # Now that we've seen a named item, no more positional.
            _parse_equals(task, cmdline, params)

        elif positional:
            # This might be a value for the next positional item.  If it also matches a
            # task name, abort though.  We are not going to guess.
            if arg in other_names:
                sys.exit('%r can be a task or a positional argument.  Use dashes or equals to disambiguate' % arg)

            params[positional[0]] = cmdline.pop(0)
            positional.pop(0)

        else:
            break


    # Now make sure all of the required arguments have a default value.  Fortunately
    # Signature.bind will do this, raising a TypeError if anything is wrong.  We'll steal
    # the message from the TypeError.

    try:
        b = task.sig.bind(**params)
        b.apply_defaults()
        return b, cmdline
    except TypeError as ex:
        sys.exit('Task %s: %s' % (task.name, ex))


def _parse_double_dash(task, cmdline, params):
    """
    Parses a "--arg" token.

    The cmdline for this argument are removed from the front of cmdline.  (The first token is
    the name of the option.  The second token may be the value.)
    """
    name  = cmdline.pop(0)[2:]
    value = inspect.Parameter.empty

    value_provided = ('=' in name)

    if value_provided:
        name, value = name.split('=', 1)

    paramname = name.replace('-', '_')

    param = task.sig.parameters.get(name)

    if not param and name.startswith('no-'):
        # If a parameter was defined as "arg=True" then we accept "--no-arg" to set the
        # value to False.
        p = task.sig.parameters.get(paramname[3:])
        if p and p.default in (True, False):
            paramname = paramname[3:]
            params[paramname] = False
            return

    if not param:
        sys.exit('Task %s does not accept an argument %r' % (task.name, name))

    if value_provided and param.default in (True, False):
        sys.exit('Task %s argument %s does not accept a value' % (task.name, name))

    if value is inspect.Parameter.empty:
        # An '=' was not provided, so we need to determine if (1) this is a Boolean
        # parameter or (2) if the next token is the value.  (We've already handled the
        # --no-xxx version.)

        if param.default is False:
            value = True
        elif param.default is True:
            # This isn't necessary, but we'll take it since it may help scripts.  The user has
            # set the default to True for something, but supplied the same flag again.
            value = True
        else:
            # The next token needs to be the value.
            if not cmdline:
                sys.exit('Task %s argument %s was not provided a value' % (task.name, name))
            value = cmdline.pop(0)

    if name in 'args':
        sys.exit('Task %s argument %s parameter was specified more than once.' % (task.name, name))

    params[paramname] = value


def _parse_equals(task, cmdlline, params):
    """
    Parse a "name=value" arg token.

    The tokens for this argument are removed from the front of cmdlline.  (The first token is
    the name of the option.  The second token may be the value.)
    """
    name, value  = cmdlline.pop(0).split('=', 1)

    paramname = name.replace('-', '_')
    param = task.sig.parameters.get(name)

    if not param:
        sys.exit('Task %s does not accept an argument %r' % (task.name, name))

    if value and param.default in (True, False):
        sys.exit('Task %s argument %s does not accept a value' % (task.name, name))

    params[paramname] = value


def _parse_flags(task, cmdline, params):
    """
    Parse "-fx value" type arguments and write their values into `params`.

    Removes parsed tokens from `cmdline`.
    """
    # Each flag is a single-letter.
    #
    # For each flag, make sure we recognize it.  (The allowed flags have already been
    # determined by the task.)
    #
    # If the flag matches a Boolean parameter, the value is True.
    #
    #   -xyz   # Three flags all set to True
    #
    # Otherwise the flag requires a parameter and must be last in the argument or must be
    # followed by an equals sign:
    #
    #   -w=Bob
    #   -w Bob
    #
    #   -xyzw=Bob # Three flags plus -w set to Bob
    #   -xyzw Bob # Three flags plus -w set to Bob

    flags = cmdline.pop(0)[1:] # remove the leading dash

    for i, flag in enumerate(flags):
        name = task.flags.get(flag)
        if not name:
            sys.exit('Task %s does not accept a -%r flag' % (task.name, flag))

        param = task.sig.parameters[name]
        if param.default in (True, False):
            params[name] = True
            continue

        # This requires a value, so it must either be last so we use the next cmdline token or
        # it must be followed by '='.

        if i == len(flags)-1:
            if not cmdline:
                sys.exit('Task %s -%r flag was not provided a value' % (task.name, flag))
            params[name] = cmdline.pop(0)
            return

        if flags[i+1] != '=':
            sys.exit('Task %s -%r flag was not provided a value' % (task.name, flag))

        params[name] = flags[i+2:]
        return




import sys, os
from os.path import dirname, join, exists

MAXDIRS  = 30
# (This is here in case I screw something up.  Better than 100% CPU)

def findfile(filename):
    path = os.getcwd()

    for i in range(MAXDIRS):
        fqn = join(path, filename)

        if exists(fqn):
            return fqn

        parent = dirname(path)
        if parent == path:
            break
        path = parent

    sys.exit('Did not find {!r}'.format(filename))




import inspect
from collections import defaultdict

_tasks = {}
# All defined tasks, mapping from function name to Task instance.

class Task:
    def __init__(self, func, flags=None):
        self.name = func.__name__
        self.func = func
        self.sig  = inspect.signature(func)

        self.flags = self._allocate_flags(flags)
        # Maps from shortflags to parameter names.

    def _allocate_flags(self, flags):
        """
        Assign flags based on the first letter of each argument.

        flags
          An optional mapping provided by the user.
        """
        final = {}
        if flags:
            final.update(flags)

        reassigned = set(final.values())

        d = defaultdict(list)
        for name in self.sig.parameters:
            # If the parameter has been assigned a different flag, don't also assign it to its
            # first letter.  This also allows us to use the letter for another flag.
            if name not in reassigned and name[0] not in final:
                d[name[0]].append(name)

        for flag, names in d.items():
            if len(names) == 1:
                final[flag] = names[0]

        return final

    def call(self, boundargs):
        """
        Calls the task, passing the arguments returned from parse.
        """
        self.func(*boundargs.args, **boundargs.kwargs)


    def __repr__(self):
        return '<Task %s>' % (self.name)


def optional_args(fn):
    # Python's decorators act completely different when called with and without arguments.
    def wrapped_decorator(*args, **kwargs):
        if len(args) == 1 and callable(args[0]):
            return fn(args[0])
        else:
            def real_decorator(decoratee):
                return fn(decoratee, *args, **kwargs)
            return real_decorator
    return wrapped_decorator


@optional_args
def task(func, flags=None):
    """
    Registers a function as a task.

    flags
      An optional mapping of single-character flags to parameter names.

    """
    task = Task(func, flags=flags)

    if task.name in _tasks:
        sys.exit('There is more than one task named {!r}'.format(task.name))
    _tasks[task.name] = task

    # Mark the function to make it easy.
    func._task = task

    return func


def get_task_names():
    """
    Returns a list of defined names.
    """
    return list(_tasks.keys())




import unittest, logging
from collections import namedtuple

from runtasks import task
from runtasks.tasks import get_task_names
from runtasks.parser import parse, parse_task

logger = logging.getLogger()

@task
def basic(arg1, arg2=False, arg3=True, arg4=None):
    pass

@task
def other(arg1=None, flag=False, forward=False, verbose=False):
    pass

@task(flags={'x': 'flag'})
def flagtest(arg1=None, flag=False, forward=False, verbose=False):
    pass

basicresults    = namedtuple('basicresults', 'arg1 arg2 arg3 arg4')
otherresults    = namedtuple('otherresults', 'arg1 flag forward verbose')
flagtestresults = namedtuple('flagtestresults', 'arg1 flag forward verbose')

class TestParser(unittest.TestCase):

    def test_positional(self):
        "Ensure position parameters work with no dashes"

        # The value "test" should be assigned to arg1 since it has no default.

        cmdline  = 'basic test'
        expected = basicresults(arg1='test', arg2=False, arg3=True, arg4=None)
        args, remaining = parse_task(basic._task, cmdline.split(), other_names=get_task_names())
        self.assertEqual(args.args, expected)
        self.assertFalse(remaining)

    def test_required(self):
        "Fail if not enough values are provided"

        # If we don't provide a value for arg1 it should fail.

        with self.assertRaises(SystemExit):
            cmdline = 'basic'
            parse_task(basic._task, cmdline.split(), other_names=get_task_names())

    def test_reject_bool_vals(self):
        "Fail if a value is provided for a Boolean parameter"

        # arg2 and arg3 both are Booleans, so you can provide "--arg2" or "--no-arg3" but not
        # "--arg2=xyz".

        with self.assertRaises(SystemExit):
            cmdline = 'basic arg2=test'
            parse_task(basic._task, cmdline.split(), other_names=get_task_names())

        with self.assertRaises(SystemExit):
            cmdline = 'basic arg3=test'
            parse_task(basic._task, cmdline.split(), other_names=get_task_names())

    def test_equals(self):
        "test arg=val"
        cmdline  = 'basic test1 arg4=test4'
        expected = basicresults(arg1='test1', arg2=False, arg3=True, arg4='test4')
        args, remaining = parse_task(basic._task, cmdline.split(), other_names=get_task_names())
        self.assertEqual(args.args, expected)
        self.assertFalse(remaining)

    def test_dashes_equals(self):
        "test --arg=val"
        cmdline  = 'basic test1 --arg4=test4'
        expected = basicresults(arg1='test1', arg2=False, arg3=True, arg4='test4')
        args, remaining = parse_task(basic._task, cmdline.split(), other_names=get_task_names())
        self.assertEqual(args.args, expected)
        self.assertFalse(remaining)

    def test_bool(self):
        "test --arg"
        # A parameter with "=False" is a Boolean parameter that can be enabled using the
        # parameter name as a flag: --arg2

        cmdline  = 'basic test1 --arg2'
        expected = basicresults(arg1='test1', arg2=True, arg3=True, arg4=None)
        args, remaining = parse_task(basic._task, cmdline.split(), other_names=get_task_names())
        self.assertEqual(args.args, expected)
        self.assertFalse(remaining)

    def test_bool_no(self):
        "Test --no-arg"
        # A parameter with "=True" is a Boolean parameter that can be disabled by prepending
        # "no-"to the parameter name as a flag: --no-arg3
        cmdline  = 'basic test1 --no-arg3'
        expected = basicresults(arg1='test1', arg2=False, arg3=False, arg4=None)
        args, remaining = parse_task(basic._task, cmdline.split(), other_names=get_task_names())
        self.assertEqual(args.args, expected)
        self.assertFalse(remaining)

    def test_bool_true_flag(self):
        "Test --arg for a default=True"
        # A parameter with "=True" is a Boolean parameter that can be disabled by prepending
        # "no-"to the parameter name as a flag: --no-arg3
        cmdline  = 'basic test1 --arg3'
        expected = basicresults(arg1='test1', arg2=False, arg3=True, arg4=None)
        args, remaining = parse_task(basic._task, cmdline.split(), other_names=get_task_names())
        self.assertEqual(args.args, expected)
        self.assertFalse(remaining)

    def test_bool_false_no(self):
        "Test --no-arg for a default=False"

        # The parameter defaults to False already, but we'll accept --no-arg for readability.
        cmdline  = 'basic test1 --no-arg2'
        expected = basicresults(arg1='test1', arg2=False, arg3=True, arg4=None)
        args, remaining = parse_task(basic._task, cmdline.split(), other_names=get_task_names())
        self.assertEqual(args.args, expected)
        self.assertFalse(remaining)

    def test_remaining(self):
        "Ensure the parser returns the remaining, unrecognized values."
        cmdline = 'basic test1 other'
        expected = basicresults(arg1='test1', arg2=False, arg3=True, arg4=None)
        args, remaining = parse_task(basic._task, cmdline.split(), other_names=get_task_names())
        self.assertEqual(args.args, expected)
        self.assertEqual(remaining, ['other'])

    def test_dont_guess(self):
        "Fail if a token is ambiguous"

        # In this case "other" can either be the value for basic.arg1 or the name of the next
        # task.  Resist the urge to guess.

        with self.assertRaises(SystemExit):
            cmdline = 'basic other'
            parse_task(basic._task, cmdline.split(), other_names=get_task_names())

        # Here's what you'd need to use if you wanted the value "other"
        cmdline  = 'basic arg1=other'
        expected = basicresults(arg1='other', arg2=False, arg3=True, arg4=None)
        args, remaining = parse_task(basic._task, cmdline.split(), other_names=get_task_names())
        self.assertEqual(args.args, expected)
        self.assertFalse(remaining)


class TestShortFlags(unittest.TestCase):
    def test_one_value(self):
        cmdline = 'other -a value1'
        expected = otherresults(arg1='value1', flag=False, forward=False, verbose=False)
        args, remaining = parse_task(other._task, cmdline.split(), other_names=get_task_names())
        self.assertEqual(args.args, expected)
        self.assertFalse(remaining)

    def test_multiple_flags(self):
        cmdline = 'other -va value1'
        expected = otherresults(arg1='value1', flag=False, forward=False, verbose=True)
        args, remaining = parse_task(other._task, cmdline.split(), other_names=get_task_names())
        self.assertEqual(args.args, expected)
        self.assertFalse(remaining)

    def test_flag(self):
        cmdline = 'other -v'
        expected = otherresults(arg1=None, flag=False, forward=False, verbose=True)
        args, remaining = parse_task(other._task, cmdline.split(), other_names=get_task_names())
        self.assertEqual(args.args, expected)
        self.assertFalse(remaining)

    def test_multiple_matches(self):
        "There can only be one parameter per letter"

        # The "other" task has two parameters that start with 'f', so 'f' is not a valid flag.

        with self.assertRaises(SystemExit):
            cmdline = 'other -f'
            parse_task(other._task, cmdline.split(), other_names=get_task_names())

    def test_manual_flags(self):
        # The flagtest task has a flag set specifically by the user.
        cmdline = 'flagtest -vx'
        expected = flagtestresults(arg1=None, flag=True, forward=False, verbose=True)
        args, remaining = parse_task(flagtest._task, cmdline.split(), other_names=get_task_names())
        self.assertEqual(args.args, expected)
        self.assertFalse(remaining)

    def test_manual_flags_multiple(self):
        # The flagtest task has two arguments that start with "f", but we've assigned "x" to
        # one of them.  That means "f" should work for the other.
        cmdline = 'flagtest -f'
        expected = flagtestresults(arg1=None, flag=False, forward=True, verbose=False)
        args, remaining = parse_task(flagtest._task, cmdline.split(), other_names=get_task_names())
        self.assertEqual(args.args, expected)
        self.assertFalse(remaining)




from distutils.cmd import Command
from distutils.core import setup


class TestCommand(Command):
    user_options = []

    def initialize_options(self):
        pass

    def finalize_options(self):
        pass

    def run(self):
        import sys, subprocess
        cmdline = '-m unittest discover -s tests -v'
        sys.exit(subprocess.call([sys.executable] + cmdline.split()))


def _getversion():
    """
    The version is maintained in runtasks/version.py.
    """
    m = __import__('runtasks.version')
    return m.version


setup(
    name='runtasks',
    description='A simple task runner for Python',
    version=_getversion(),
    author='Michael Kleehammer',
    author_email='michael@kleehammer.com',
    url='https://gitlab.com/mkleehammer/runtasks',
    packages=['runtasks'],
    scripts=['scripts/run', 'scripts/run.cmd'],
    classifiers=[
        'Development Status :: 5 - Production/Stable',
        'Environment :: Console',
        'Intended Audience :: Developers',
        'Intended Audience :: System Administrators',
        'License :: OSI Approved :: MIT License',
        'Operating System :: MacOS :: MacOS X',
        'Operating System :: Microsoft :: Windows',
        'Operating System :: POSIX',
        'Programming Language :: Python',
        'Topic :: Software Development :: Build Tools'
    ],
    cmdclass=dict(test=TestCommand)
)
